"""
This is module should not import any other runez module, it's the lowest on the import chain
"""

import os
import re


SYMBOLIC_TMP = "<tmp>"
RE_FORMAT_MARKERS = re.compile(r"{([^}]*?)}")


def flattened(value, separator=None, unique=True):
    """
    :param value: Possibly nested arguments (sequence of lists, nested lists)
    :param str|None separator: Split values with 'separator' if specified
    :param bool unique: If True, return unique values only
    :return list: 'value' flattened out (leaves from all involved lists/tuples)
    """
    result = []
    _flatten(result, value, separator=separator, unique=unique)
    return result


def formatted(text, *args, **kwargs):
    """
    :param str text: Text to format
    :param args: Objects to extract values from (as attributes)
    :param kwargs: Optional values provided as named args
    :return str: Attributes from this class are expanded if mentioned
    """
    objects = list(args) + [kwargs] if kwargs else args[0] if len(args) == 1 else args
    if not text or not objects:
        return text
    values = {}
    markers = RE_FORMAT_MARKERS.findall(text)
    while markers:
        key = markers.pop()
        if key in values:
            continue
        val = _find_value(key, objects)
        if val is None:
            return None
        val = str(val)
        markers.extend(m for m in RE_FORMAT_MARKERS.findall(val) if m not in values)
        values[key] = val
    for key, val in values.items():
        if '{' in val:
            values[key] = values[key].format(**values)
    return text.format(**values)


def quoted(text):
    """
    :param str|None text: Text to optionally quote
    :return str: Quoted if 'text' contains spaces
    """
    if text and " " in text:
        sep = "'" if '"' in text else '"'
        return "%s%s%s" % (sep, text, sep)
    return text


def represented_args(args, separator=" "):
    """
    :param list|tuple|None args: Arguments to represent
    :param str separator: Separator to use
    :return str: Quoted as needed textual representation
    """
    result = []
    if args:
        for text in args:
            result.append(quoted(short(text)))
    return separator.join(result)


def resolved_path(path, base=None):
    """
    :param str|None path: Path to resolve
    :param str|None base: Base path to use to resolve relative paths (default: current working dir)
    :return str: Absolute path
    """
    if not path or path.startswith(SYMBOLIC_TMP):
        return path

    path = os.path.expanduser(path)
    if base and not os.path.isabs(path):
        return os.path.join(resolved_path(base), path)

    return os.path.abspath(path)


def short(path):
    return Anchored.short(path)


def shortened(text, size=120):
    """
    :param str text: Text to shorten
    :param int size: Max chars
    :return str: Leading part of 'text' with at most 'size' chars
    """
    if text:
        text = text.strip()
        if len(text) > size:
            return "%s..." % text[:size - 3].strip()
    return text


def to_int(text, default=None):
    """
    :param text: Value to convert
    :param int|None default: Default to use if 'text' can't be parsed
    :return int:
    """
    try:
        return int(text)

    except (TypeError, ValueError):
        return default


class Anchored:
    """
    An "anchor" is a known path that we don't wish to show in full when printing/logging
    This allows to conveniently shorten paths, and show more readable relative paths
    """

    paths = []  # Folder paths that can be used to shorten paths, via short()
    home = os.path.expanduser("~")

    def __init__(self, folder):
        self.folder = resolved_path(folder)

    def __enter__(self):
        Anchored.add(self.folder)

    def __exit__(self, *_):
        Anchored.pop(self.folder)

    @classmethod
    def set(cls, *anchors):
        """
        :param str|list anchors: Optional paths to use as anchors for short()
        """
        cls.paths = sorted(flattened(anchors, unique=True), reverse=True)

    @classmethod
    def add(cls, anchors):
        """
        :param str|list anchors: Optional paths to use as anchors for short()
        """
        cls.set(cls.paths, anchors)

    @classmethod
    def pop(cls, anchors):
        """
        :param str|list anchors: Optional paths to use as anchors for short()
        """
        for anchor in flattened(anchors):
            if anchor in cls.paths:
                cls.paths.remove(anchor)

    @classmethod
    def short(cls, path):
        """
        Example:
            short("examined /Users/joe/foo") => "examined ~/foo"

        :param path: Path to represent in its short form
        :return str: Short form, using '~' if applicable
        """
        if not path:
            return path

        path = str(path)
        if cls.paths:
            for p in cls.paths:
                if p:
                    path = path.replace(p + "/", "")

        path = path.replace(cls.home, "~")
        return path


def _flatten(result, value, separator=None, unique=True):
    """
    :param list result: Will hold all flattened values
    :param value: Possibly nested arguments (sequence of lists, nested lists)
    :param str|None separator: Split values with 'separator' if specified
    :param bool unique: If True, return unique values only
    """
    if not value:
        # Convenience: allow to filter out --foo None easily
        if value is None and not unique and result and result[-1].startswith("-"):
            result.pop(-1)
        return

    if isinstance(value, (list, tuple, set)):
        for item in value:
            _flatten(result, item, separator=separator, unique=unique)
        return

    if separator is not None and hasattr(value, "split") and separator in value:
        _flatten(result, value.split(separator), separator=separator, unique=unique)
        return

    if not unique or value not in result:
        result.append(value)


def _get_value(obj, key):
    """Get a value for 'key' from 'obj', if possible"""
    if isinstance(obj, (list, tuple)):
        for item in obj:
            v = _find_value(key, item)
            if v is not None:
                return v
        return None
    if isinstance(obj, dict):
        return obj.get(key)
    if obj is not None:
        return getattr(obj, key, None)


def _find_value(key, *args):
    """Find a value for 'key' in any of the objects given as 'args'"""
    for arg in args:
        v = _get_value(arg, key)
        if v is not None:
            return v



"""
Convenience context managers
"""

import os
import shutil
import sys
import tempfile

try:
    import StringIO
    StringIO = StringIO.StringIO

except ImportError:
    from io import StringIO

from runez.convert import Anchored, resolved_path, SYMBOLIC_TMP
from runez.logsetup import LogManager
from runez.system import AbortException, is_dryrun, set_dryrun


class CapturedStream:
    """Capture output to a stream by hijacking temporarily its write() function"""

    _shared = None

    def __init__(self, name, target=None, buffer=None):
        self.name = name
        self.target = target
        if buffer is not None:
            self.buffer = StringIO(buffer.getvalue())
            self.name += "*"

        elif target is None:
            self.buffer = CapturedStream._shared._buffer
        else:
            self.buffer = StringIO()

    @classmethod
    def log_intercept(cls):
        if cls._shared:
            return cls("log")

    def __repr__(self):
        return self.contents()

    def __eq__(self, other):
        if isinstance(other, CapturedStream):
            return self.name == other.name and self.contents() == other.contents()
        return str(self).strip() == str(other).strip()

    def __contains__(self, item):
        return item is not None and item in self.contents()

    def __len__(self):
        return len(self.contents())

    def duplicate(self):
        return CapturedStream(self.name, buffer=StringIO(self.contents()))

    def contents(self):
        return self.buffer.getvalue()

    def write(self, message):
        self.buffer.write(message)

    def capture(self):
        if self.target:
            self.original = self.target.write
            self.target.write = self.buffer.write
        elif self.name == "log":
            self._shared._is_capturing = True

    def restore(self):
        """Restore hijacked write() function"""
        if self.target:
            self.target.write = self.original
        elif self.name == "log":
            self._shared._is_capturing = False

    def pop(self):
        """Current content popped, useful for testing"""
        r = self.contents()
        self.clear()
        return r

    def clear(self):
        """Clear captured content"""
        self.buffer.seek(0)
        self.buffer.truncate(0)


def _dupe(cap):
    """
    :param CapturedStream|None cap:
    :return CapturedStream|None: Duplicate of 'cap', if any
    """
    if cap is not None:
        return cap.duplicate()


class TrackedOutput(object):
    """
    Track captured output
    """

    def __init__(self, stdout, stderr, log):
        """
        :param CapturedStream|None stdout: Captured stdout
        :param CapturedStream|None stderr: Captured stderr
        :param CapturedStream|None log: Captured logging
        """
        self.stdout = stdout
        self.stderr = stderr
        self.log = log
        self.captured = [c for c in (self.stdout, self.stderr, self.log) if c is not None]

    def __repr__(self):
        return "\n".join("%s: %s" % (s.name, s) for s in self.captured)

    def __eq__(self, other):
        if isinstance(other, CaptureOutput):
            return self.stdout == other.stdout and self.stderr == other.stderr and self.log == other.log
        return str(self).strip() == str(other).strip()

    def __contains__(self, item):
        return any(item in s for s in self.captured)

    def __len__(self):
        return sum(len(s) for s in self.captured)

    def duplicate(self):
        return TrackedOutput(_dupe(self.stdout), _dupe(self.stderr), _dupe(self.log))

    def contents(self):
        return "".join(s.contents() for s in self.captured)

    def pop(self):
        """Current content popped, useful for testing"""
        r = self.contents()
        self.clear()
        return r

    def clear(self):
        """Clear captured content"""
        for s in self.captured:
            s.clear()


class CaptureOutput(TrackedOutput):
    """
    Context manager allowing to temporarily grab stdout/stderr output.
    Output is captured and made available only for the duration of the context.

    Sample usage:

    with CaptureOutput() as logged:
        # do something that generates output
        # output has been captured in 'logged'
    """

    def __init__(self, level=None, stdout=True, stderr=True, log=None, anchors=None, dryrun=None):
        """
        :param int|None level: Change logging level, if specified
        :param bool stdout: Capture stdout
        :param bool stderr: Capture stderr
        :param bool|None log: Capture pytest logging (if running in pytest), leave at None for auto-detect
        :param str|list anchors: Optional paths to use as anchors for short()
        :param bool|None dryrun: Override dryrun (when explicitly specified, ie not None)
        """
        stdout = CapturedStream("stdout", target=sys.stdout) if stdout else None
        stderr = CapturedStream("stderr", target=sys.stderr) if stderr else None
        if log is None:
            log = bool(CapturedStream._shared)
        log = CapturedStream.log_intercept() if log else None
        super(CaptureOutput, self).__init__(stdout, stderr, log)
        self.level = level
        self.anchors = anchors
        self.dryrun = dryrun
        self._old_level = None

    def __enter__(self):
        self._old_level = LogManager.override_root_level(self.level)
        for s in self.captured:
            s.capture()
        if self.anchors:
            Anchored.add(self.anchors)
        if self.dryrun is not None:
            self.dryrun = set_dryrun(self.dryrun)
        return self

    def __exit__(self, *args):
        LogManager.override_root_level(self._old_level)
        for s in self.captured:
            s.restore()
        if self.anchors:
            Anchored.pop(self.anchors)
        if self.dryrun is not None:
            set_dryrun(self.dryrun)
        self.clear()


class CurrentFolder:
    """
    Context manager for changing the current working directory
    """

    def __init__(self, destination, anchor=False):
        self.anchor = anchor
        self.destination = resolved_path(destination)

    def __enter__(self):
        self.current_folder = os.getcwd()
        os.chdir(self.destination)
        if self.anchor:
            Anchored.add(self.destination)

    def __exit__(self, *_):
        os.chdir(self.current_folder)
        if self.anchor:
            Anchored.pop(self.destination)


class TempFolder:
    """
    Context manager for obtaining a temp folder
    """

    def __init__(self, anchor=True, dryrun=None, follow=True):
        """
        :param anchor: If True, short-ify paths relative to used temp folder
        :param dryrun: Override dryrun (if provided)
        :param follow: If True, change working dir to temp folder (and restore)
        """
        self.anchor = anchor
        self.dryrun = dryrun
        self.follow = follow
        self.old_cwd = None
        self.tmp_folder = None

    def __enter__(self):
        if self.dryrun is not None:
            self.dryrun = set_dryrun(self.dryrun)
        if not is_dryrun():
            # Use realpath() to properly resolve for example symlinks on OSX temp paths
            self.tmp_folder = os.path.realpath(tempfile.mkdtemp())
            if self.follow:
                self.old_cwd = os.getcwd()
                os.chdir(self.tmp_folder)
        tmp = self.tmp_folder or SYMBOLIC_TMP
        if self.anchor:
            Anchored.add(tmp)
        return tmp

    def __exit__(self, *_):
        if self.anchor:
            Anchored.pop(self.tmp_folder or SYMBOLIC_TMP)
        if self.old_cwd:
            os.chdir(self.old_cwd)
        if self.tmp_folder:
            shutil.rmtree(self.tmp_folder)
        if self.dryrun is not None:
            set_dryrun(self.dryrun)


def verify_abort(func, *args, **kwargs):
    """
    Convenient wrapper around functions that should exit or raise an exception

    Example:
        assert "Can't create folder" in verify_abort(ensure_folder, "/dev/null/not-there")

    :param callable func: Function to execute
    :param args: Args to pass to 'func'
    :param Exception expected_exception: Type of exception that should be raised
    :param kwargs: Named args to pass to 'func'
    :return str: Chatter from call to 'func', if it did indeed raise
    """
    expected_exception = kwargs.pop("expected_exception", AbortException)
    with CaptureOutput() as logged:
        try:
            func(*args, **kwargs)
            return None
        except expected_exception:
            return str(logged)



import logging
import time


class AbortException(Exception):
    """
    You can replace this with your preferred exception, for example:

        runez.base.AbortException = SystemExit
    """
    def __init__(self, code):
        self.code = code


def abort(*args, **kwargs):
    """
    Usage:
        return abort("...") => will sys.exit() by default
        return abort("...", fatal=True) => Will sys.exit()

        # Not fatal, but will log/print message:
        return abort("...", fatal=False) => Will return False
        return abort("...", fatal=(False, None)) => Will return None
        return abort("...", fatal=(False, -1)) => Will return -1

        # Not fatal, will not log/print any message:
        return abort("...", fatal=None) => Will return None
        return abort("...", fatal=(None, None)) => Will return None
        return abort("...", fatal=(None, -1)) => Will return -1

    :param args: Args passed through for error reporting
    :param kwargs: Args passed through for error reporting
    :return: kwargs["return_value"] (default: -1) to signify failure to non-fatal callers
    """
    code = kwargs.pop("code", 1)
    logger = kwargs.pop("logger", logging.error if code else logging.info)
    fatal = kwargs.pop("fatal", True)
    return_value = fatal
    if isinstance(fatal, tuple) and len(fatal) == 2:
        fatal, return_value = fatal
    if logger and fatal is not None and args:
        logger(*args, **kwargs)
    if fatal:
        if isinstance(fatal, type) and issubclass(fatal, BaseException):
            raise fatal(code)
        if AbortException is not None:
            if isinstance(AbortException, type) and issubclass(AbortException, BaseException):
                raise AbortException(code)
            return AbortException(code)
    return return_value


def get_timezone():
    try:
        return time.tzname[0]
    except (IndexError, TypeError):
        return ""


def get_version(mod, default="0.0.0"):
    """
    :param module|str mod: Module, or module name to find version for (pass either calling module, or its .__name__)
    :param str default: Value to return if version determination fails
    :return str: Determined version
    """
    name = mod
    if hasattr(mod, "__name__"):
        name = mod.__name__

    try:
        import pkg_resources
        return pkg_resources.get_distribution(name).version

    except Exception as e:
        logging.warning("Can't determine version for %s: %s", name, e, exc_info=e)
        return default


# We have to import 'runez' late, can't import it right away at import time
_runez_module = None


def _get_runez():
    global _runez_module
    if _runez_module is None:
        import runez
        _runez_module = runez
    return _runez_module


def is_dryrun():
    return _get_runez().DRYRUN


def set_dryrun(dryrun):
    """
    :param bool dryrun: New value for runez.DRYRUN
    :return bool: Old value
    """
    r = _get_runez()
    old = r.DRYRUN
    r.DRYRUN = bool(dryrun)
    return old



"""
Base functionality used by other parts of runez

We track here whether we're running in dryrun mode, convenience logging etc
"""

import inspect
import threading


try:
    string_type = basestring  # noqa

except NameError:
    string_type = str
    unicode = str


def decode(value):
    """Python 2/3 friendly decoding of output"""
    if isinstance(value, bytes) and not isinstance(value, str):
        return value.decode("utf-8")
    return unicode(value)


class prop(object):
    """
    Decorator for settable cached properties.
    This comes in handy for properties you'd like to avoid computing multiple times,
    yet be able to arbitrarily change them as well, and be able to know when they get changed.
    """

    def __init__(self, func):
        """
        :param callable: Wrapped function
        """
        self.function = func
        self.name = func.__name__
        self.field_name = "__%s" % self.name
        self.on_prop, self.prop_arg = _find_on_prop(inspect.currentframe().f_back)
        self.__doc__ = func.__doc__

    def __repr__(self):
        return self.name

    def __get__(self, instance, cls=None):
        if instance is None:
            instance = cls
        cached = getattr(instance, self.field_name, None)
        if cached is None:
            cached = self.function(instance)
            setattr(instance, self.field_name, cached)
        return cached

    def __set__(self, instance, value):
        setattr(instance, self.field_name, value)
        if self.on_prop:
            if self.prop_arg:
                self.on_prop(instance, prop=self)
            else:
                self.on_prop(instance)


class Slotted(object):
    """This class allows to easily initialize/set a descendant using named arguments"""

    def __init__(self, *args, **kwargs):
        """
        :param args: Optionally provide another instance of same type to initialize from
        :param kwargs: Override one or more of this classes' fields listed above
        """
        self._seed()
        self.set(*args, **kwargs)

    def _seed(self):
        """Seed initial fields"""
        for name in self.__slots__:
            setattr(self, name, None)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            for name in self.__slots__:
                if getattr(self, name, None) != getattr(other, name, None):
                    return False
            return True

    def _set(self, name, value):
        setattr(self, name, value)

    def set(self, *args, **kwargs):
        """Conveniently set one or more fields at a time"""
        if args:
            if kwargs:
                raise ValueError("Provide either one positional, or field values as named arguments, but not both")
            if len(args) > 1:
                raise ValueError("Provide only one other %s as positional argument" % self.__class__.__name__)
            other = args[0]
            if isinstance(other, self.__class__):
                for name in self.__slots__:
                    self._set(name, getattr(other, name))
                return
            raise ValueError("Argument is not of type %s: %s" % (self.__class__.__name__, args[0]))
        for name in kwargs:
            if not hasattr(self, name):
                raise ValueError("Unknown %s field '%s'" % (self.__class__.__name__, name))
            self._set(name, kwargs[name])


class ThreadGlobalContext:
    """
    Thread-local + global context, composed of key/value pairs
    Thread-local context is a dict per thread (stored in a threading.local())
    Global context is a simple dict (applies to all threads)
    """

    def __init__(self, filter_type):
        """
        :param type filter_type: Class to instantiate as filter
        """
        self._filter_type = filter_type
        self._lock = threading.RLock()
        self._tpayload = None
        self._gpayload = None
        self.filter = None

    def reset(self):
        with self._lock:
            self.filter = None
            self._tpayload = None
            self._gpayload = None

    def enable(self):
        """Enable contextual logging"""
        with self._lock:
            if self.filter is None:
                self.filter = self._filter_type(self)

    def has_threadlocal(self):
        with self._lock:
            return bool(self._tpayload)

    def has_global(self):
        with self._lock:
            return bool(self._gpayload)

    def set_threadlocal(self, **values):
        """Set current thread's logging context to 'values'"""
        with self._lock:
            self._ensure_threadlocal()
            self._tpayload.context = values

    def add_threadlocal(self, **values):
        """Add 'values' to current thread's logging context"""
        with self._lock:
            self._ensure_threadlocal()
            self._tpayload.context.update(**values)

    def remove_threadlocal(self, name):
        """Remove entry with 'name' from current thread's context"""
        with self._lock:
            if self._tpayload is not None:
                if name in self._tpayload.context:
                    del self._tpayload.context[name]
                if not self._tpayload.context:
                    self._tpayload = None

    def clear_threadlocal(self):
        """Clear current thread's context"""
        with self._lock:
            self._tpayload = None

    def set_global(self, **values):
        """Set global logging context to 'values'"""
        with self._lock:
            self._ensure_global(values)

    def add_global(self, **values):
        """Add 'values' to global logging context"""
        with self._lock:
            self._ensure_global()
            self._gpayload.update(**values)

    def remove_global(self, name):
        """Remove entry with 'name' from global context"""
        with self._lock:
            if self._gpayload is not None:
                if name in self._gpayload:
                    del self._gpayload[name]
                if not self._gpayload:
                    self._gpayload = None

    def clear_global(self):
        """Clear global context"""
        with self._lock:
            if self._gpayload is not None:
                self._gpayload = None

    def to_dict(self):
        """
        :return dict: Combined global and thread-specific logging context
        """
        with self._lock:
            result = {}
            if self._gpayload:
                result.update(self._gpayload)
            if self._tpayload:
                result.update(getattr(self._tpayload, "context", {}))
            return result

    def _ensure_threadlocal(self):
        if self._tpayload is None:
            self._tpayload = threading.local()
            self._tpayload.context = {}
        self.enable()

    def _ensure_global(self, values=None):
        if self._gpayload is None:
            self._gpayload = values or {}
        self.enable()


if hasattr(inspect, "signature"):
    # python3
    def _has_arg(func, arg_name):
        return arg_name in inspect.signature(func).parameters

else:
    # python2
    def _has_arg(func, arg_name):
        return arg_name in inspect.getargspec(func).args


def _find_on_prop(frame):
    """
    :param frame frame: Frame to examine
    :return function|None, bool: __on_prop function if any, boolean indicates whether that function takes 'prop' as argument
    """
    for name, func in frame.f_locals.items():
        if name.endswith("__on_prop"):
            if isinstance(func, classmethod):
                func = func.__func__
            return func, _has_arg(func, "prop")
    return None, False



"""
Convenience logging setup
"""

import inspect
import logging
import os
import signal
import sys
import threading
from functools import partial

try:
    # faulthandler is only available in python 3.3+
    import faulthandler

except ImportError:
    faulthandler = None

import runez.system
from runez.base import Slotted, ThreadGlobalContext
from runez.convert import flattened, formatted
from runez.path import basename as get_basename, resolved_location
from runez.program import get_dev_folder, get_program_path


ORIGINAL_CF = logging.currentframe


class LoggingSnapshot(Slotted):
    """
    Take a snapshot of parts we're modifying in the 'logging' module, in order to be able to restore it as it was
    """

    __slots__ = ["_srcfile", "critical", "fatal", "error", "exception", "warning", "info", "debug"]

    def _seed(self):
        """Seed initial fields"""
        for name in self.__slots__:
            setattr(self, name, getattr(logging, name))

    def restore(self):
        for name in self.__slots__:
            setattr(logging, name, getattr(self, name))


class LogSpec(Slotted):
    """
    Settings to use, you can safely customize these (before calling runez.log.setup), for example:
        runez.log.setup(console_stream=sys.stdout)
    or
        runez.log.spec.set(console_stream=sys.stdout)
    or
        runez.log.spec.console_stream = sys.stdout
    """

    __slots__ = [
        "appname", "basename", "level", "console_stream", "locations", "rotate",
        "console_format", "file_format", "context_format", "timezone", "dev", "tmp", "greeting",
    ]

    def resolved_location(self, custom_location):
        """
        :param str custom_location: Custom location to use (if not None)
        :return str|None: First available location
        """
        return resolved_location(self, custom_location, locations=self.locations, basename=self.basename)


class _ContextFilter(logging.Filter):
    """
    Optional logging filter allowing to inject key/value pairs to every log record.

    In order to activate this:
    - Mention %(context)s in your log format
    - Add key/value pairs via runez.log.context.add_global(), runez.log.context.add_threadlocal()
    """

    def __init__(self, context, name=""):
        """
        :param ThreadGlobalContext context: Associated context
        :param str name: Passed through to parent
        """
        super(_ContextFilter, self).__init__(name=name)
        self.context = context

    def filter(self, record):
        """Determines if the record should be logged and injects context info into the record. Always returns True"""
        fmt = LogManager.spec.context_format
        if fmt:
            data = self.context.to_dict()
            if data:
                record.context = fmt % ",".join("%s=%s" % (key, val) for key, val in sorted(data.items()) if key and val)
            else:
                record.context = ""
        return True


class LogManager:
    """
    Global logging context managed by runez.
    There's only one, as multiple contexts would not be useful (logging setup is a global thing)
    """

    # Defaults used to initialize LogSpec instances
    # You shouldn't need to change these, but changing prior to calling runez.log.setup() will take effect
    # Use runez.log.override_spec() to change these defaults
    _default_spec = LogSpec(
        appname=get_basename(get_program_path()),
        basename="{appname}.log",
        level=logging.INFO,
        console_stream=sys.stderr,
        locations=["{dev}/log/{basename}", "/logs/{appname}/{basename}", "/var/log/{basename}"],
        rotate=None,
        console_format="%(asctime)s %(levelname)s %(message)s",
        file_format="%(asctime)s %(timezone)s [%(threadName)s] %(context)s%(levelname)s - %(message)s",
        context_format="[[%s]] ",
        timezone=runez.system.get_timezone(),
        greeting="Pid {pid}, logging to {location}",
    )

    # Spec defines how logs should be setup()
    # Best way to provide your spec is via: runez.log.setup(), for example:
    #   runez.log.setup(appname="my-app)
    spec = LogSpec(_default_spec)

    # Thread-local / global context
    context = ThreadGlobalContext(_ContextFilter)

    # Below fields should be read-only for outside users, do not modify these
    level = None  # Current severity level
    file_handler = None  # type: logging.FileHandler # File we're currently logging to (if any)
    handlers = None  # type: list[logging.StreamHandler]
    used_formats = None  # type: str
    faulthandler = faulthandler
    faulthandler_signum = None  # type: int

    _lock = threading.RLock()
    _logging_snapshot = LoggingSnapshot()

    @classmethod
    def setup(cls, debug=None, dryrun=None, location=None, **kwargs):
        """
        :param bool|None debug: Enable debug level logging
        :param bool|None dryrun: Enable dryrun
        :param str|None location: Optional custom log file location to use
        :param kwargs: Additional settings passed-through to 'spec'
        """
        with cls._lock:
            if cls.level is not None:
                raise Exception("Please call runez.log.setup() only once")
            if dryrun is not None:
                if dryrun and debug is None:
                    debug = True
                runez.system.set_dryrun(dryrun)
            cls.spec.set(**kwargs)
            cls.level = logging.DEBUG if debug else cls.spec.level
            if logging.root.level != cls.level:
                logging.root.setLevel(cls.level)
            if cls.spec.dev is None:
                cls.spec.dev = get_dev_folder()
            hconsole = _get_handler(cls.level, logging.StreamHandler, cls.spec.console_format, cls.spec.console_stream)
            location = cls.spec.resolved_location(location)
            if location:
                cls.file_handler = _get_handler(cls.level, logging.FileHandler, cls.spec.file_format, location)
            cls.handlers = [h for h in (hconsole, cls.file_handler) if h is not None]
            cls.used_formats = " ".join(h.formatter._fmt for h in cls.handlers)

            if cls.is_using_format("%(context)s"):
                cls.context.enable()
            if cls.is_using_format("%(name)s"):
                cls._fix_logging_shortcuts()
            if cls.context.filter:
                for handler in cls.handlers:
                    handler.addFilter(cls.context.filter)

            if cls.file_handler and cls.spec.greeting:
                message = formatted(cls.spec.greeting, pid=os.getpid(), location=location)
                cls.greet(message)

    @classmethod
    def greet(cls, message, logger=None):
        """
        :param str message: Message to log
        :param logger: Logger to use (default: logging.debug)
        """
        if message:
            if logger is None:
                logger = logging.debug
            logger(message)

    @classmethod
    def silence(cls, *modules):
        """
        :param modules: Modules, or names of modules to silence, by setting their log level to WARNING
        """
        for mod in modules:
            name = mod.__name__ if hasattr(mod, "__name__") else mod
            logging.getLogger(name).setLevel(logging.WARNING)

    @classmethod
    def is_using_format(cls, markers, used_formats=None):
        """
        :param str markers: Space separated list of markers to look for
        :param str used_formats: Formats to consider (default: cls.used_formats)
        :return bool: True if any one of the 'markers' is seen in 'used_formats'
        """
        """"""
        if used_formats is None:
            used_formats = cls.used_formats
        if not markers or not used_formats:
            return False
        return any(marker in used_formats for marker in flattened(markers, separator=" ", unique=True))

    @classmethod
    def enable_faulthandler(cls, signum=signal.SIGUSR1):
        """
        Enable dumping thread stack traces when specified signals are received, similar to java's handling of SIGQUIT

        Note: this must be called from the surviving process in case of daemonization.
        Note that SIGQUIT does not work in all environments with a python process.

        :param int|None signum: Signal number to register for full thread stack dump (use None to disable)
        """
        with cls._lock:
            if not signum:
                cls._disable_faulthandler()
                return
            if not cls.file_handler or faulthandler is None:
                return
            cls.faulthandler_signum = signum
            dump_file = cls.file_handler.stream
            faulthandler.enable(file=dump_file, all_threads=True)
            faulthandler.register(signum, file=dump_file, all_threads=True, chain=False)

    @classmethod
    def override_root_level(cls, level):
        """This function is useful for testing, to temporarily set logging level to debug basically"""
        if level is not None:
            old_level = cls._default_spec.level
            if level != cls._default_spec.level:
                cls._default_spec.level = level
                cls.spec.level = level
            if level != logging.root.level:
                logging.root.setLevel(level)
            return old_level

    @classmethod
    def override_spec(cls, **kwargs):
        """OVerride 'spec' and '_default_spec' with given values"""
        cls._default_spec.set(**kwargs)
        cls.spec.set(**kwargs)

    @classmethod
    def _reset(cls):
        """Reset logging as it was before setup(), no need to call this outside of testing, or some very special cases"""
        cls._disable_faulthandler()
        if cls.handlers is not None:
            for handler in cls.handlers:
                logging.root.removeHandler(handler)
        cls._logging_snapshot.restore()
        cls.context.reset()
        cls.spec = LogSpec(cls._default_spec)
        cls.level = None
        cls.file_handler = None
        cls.handlers = None
        cls.used_formats = None

    @classmethod
    def _disable_faulthandler(cls):
        if faulthandler and cls.faulthandler_signum:
            faulthandler.unregister(cls.faulthandler_signum)
            faulthandler.disable()
            cls.faulthandler_signum = None

    @classmethod
    def _fix_logging_shortcuts(cls):
        """
        Fix standard logging shortcuts to correctly report logging module.

        This is only useful if you:
        - actually use %(name) and care about it being correct
        - you would still like to use the logging.info() etc shortcuts

        So basically you'd like to write this:
            import logging
            logging.info("hello")

        Instead of this:
            import logging
            LOG = logging.getLogger(__name__)
            LOG.info("hello")
        """
        if cls.is_using_format("%(pathname)s %(filename)s %(funcName)s %(module)s"):
            logging._srcfile = cls._logging_snapshot._srcfile
        else:
            logging._srcfile = None

        logging.logProcesses = cls.is_using_format("%(process)d")
        logging.logThreads = cls.is_using_format("%(thread)d %(threadName)s")

        def getframe():
            return sys._getframe(4)

        def log(level, msg, *args, **kwargs):
            """Wrapper to make logging.info() etc report the right module %(name)"""
            name = inspect.currentframe().f_back.f_globals["__name__"]
            logger = logging.getLogger(name)
            try:
                logging.currentframe = getframe
                logger.log(level, msg, *args, **kwargs)
            finally:
                logging.currentframe = ORIGINAL_CF

        def wrap(level, **kwargs):
            """Wrap corresponding logging shortcut function"""
            original = getattr(logging, logging.getLevelName(level).lower())
            f = partial(log, level, **kwargs)
            f.__doc__ = original.__doc__
            return f

        logging.critical = wrap(logging.CRITICAL)
        logging.fatal = logging.critical
        logging.error = wrap(logging.ERROR)
        logging.exception = partial(logging.error, exc_info=True)
        logging.warning = wrap(logging.WARNING)
        logging.info = wrap(logging.INFO)
        logging.debug = wrap(logging.DEBUG)
        logging.log = log


def _replace_and_pad(format, marker, replacement):
    """
    :param str format: Format to tweak
    :param str marker: Marker to replace
    :param str replacement: What to replace marker with
    :return str: Resulting format, with marker replaced
    """
    if marker not in format:
        return format
    if replacement:
        return format.replace(marker, replacement)
    # Remove mention of 'marker' in 'format', including leading space (if present)
    format = format.replace("%s " % marker, marker)
    return format.replace(marker, "")


def _get_handler(level, base, format, target):
    """
    :param type(logging.Handler) base: Handler implementation to use
    :param str|None format: Format to use
    :param str|None target: Target for handler
    :return logging.Handler|None:
    """
    if format and target:
        handler = base(target)
        handler.setFormatter(_get_formatter(format))
        handler.setLevel(level)
        logging.root.addHandler(handler)
        return handler


def _get_formatter(format):
    """
    :param str format: Format specification
    :return logging.Formatter: Associated logging formatter
    """
    format = _replace_and_pad(format, "%(timezone)s", LogManager.spec.timezone)
    return logging.Formatter(format)



"""
Import this only from your test cases

Example:

    from runez.conftest import cli, isolated_log_setup, temp_folder
"""

import logging
import os
import re

import _pytest.logging
import pytest

import runez
from runez.base import string_type


runez.log.override_root_level(logging.DEBUG)
runez.log.override_spec(appname="pytest", timezone="UTC", locations=["{tmp}/{basename}"], tmp=os.path.join("/", "tmp"))


class IsolatedLogs:
    """
    Allows to isolate changes to logging setup.
    This should only be useful for testing (as in general, logging setup is a global thing)
    """

    def __enter__(self):
        """Context manager to save and restore log setup, useful for testing"""
        return runez.log

    def __exit__(self, *_):
        runez.log._reset()


@pytest.fixture
def cli():
    """
    Convenience for click CLI testing.

    Example usage:

        from runez.conftest import cli
        from my_cli import main

        cli.default_main = main  # Handy if you have only one main

        def test_help(cli):
            cli.main = main  # Not needed if `cli.default_main` was set
            cli.run("--help", main=main)  # 'main' not needed if already set
            assert cli.succeeded
            assert cli.match("Usage:")
            # or more specifically
            assert "Usage:" in cli.logged.stdout

    """
    yield ClickRunner()


# Comes in handy for click apps with only one main entry point
cli.default_main = None


@pytest.fixture
def isolated_log_setup():
    """Log settings restored"""
    with runez.TempFolder(follow=True) as tmp:
        with IsolatedLogs() as isolated:
            isolated.spec.tmp = tmp
            yield isolated


@pytest.fixture
def logged():
    with runez.CaptureOutput() as logged:
        yield logged


@pytest.fixture
def temp_folder():
    with runez.TempFolder() as tmp:
        yield tmp


class WrappedHandler(_pytest.logging.LogCaptureHandler):
    """pytest aggressively imposes its own capture, this allows to capture it in our context managers"""

    _is_capturing = False
    _buffer = runez.context.StringIO()

    def __init__(self):
        super(WrappedHandler, self).__init__()

    def emit(self, record):
        if self._is_capturing:
            msg = self.format(record)
            WrappedHandler._buffer.write(msg)
        else:
            super(WrappedHandler, self).emit(record)


runez.context.CapturedStream._shared = WrappedHandler
_pytest.logging.LogCaptureHandler = WrappedHandler


class ClickWrapper:
    """Wrap click invoke, when click is available, otherwise just call provided function"""

    __runner = None

    def __init__(self, output=None, exit_code=None, exception=None):
        self.output = output
        self.exit_code = exit_code
        self.exception = exception

    def invoke(self, main, args, **kwargs):
        """Mocked click-like behavior"""
        try:
            output = main(*args, **kwargs)
            return ClickWrapper(output=output, exit_code=0)

        except Exception as e:
            return ClickWrapper(str(e), exit_code=1, exception=e)

    @runez.prop
    def runner(cls):
        """
        :return type: CliRunner if available
        """
        try:
            from click.testing import CliRunner
            return CliRunner  # pragma: no cover, click used only if installed

        except ImportError:
            return cls


class ClickRunner:
    """Allows to provide a test-friendly fixture around testing click entry-points"""

    def __init__(self):
        self.main = cli.default_main
        self.logged = None  # type: runez.TrackedOutput
        self.exit_code = None  # type: int

    def run(self, *args, **kwargs):
        """
        :param str|list args: Command line args
        """
        if len(args) == 1 and hasattr(args[0], "split"):
            # Convenience: allow to provide full command as one string argument
            args = args[0].split()
        args = runez.flattened(args, unique=False)
        cmd = kwargs.pop("main", self.main)
        assert bool(cmd), "No main provided"
        with IsolatedLogs():
            with runez.CaptureOutput(dryrun=runez.DRYRUN) as logged:
                runner = ClickWrapper.runner
                runner = runner()
                result = runner.invoke(cmd, args=args, **kwargs)
                if result.output:
                    logged.stdout.write(result.output)
                if result.exception:
                    try:
                        raise result.exception
                    except BaseException:
                        logging.exception("Exited with stacktrace:")
                self.logged = logged.duplicate()
                self.exit_code = result.exit_code

    @property
    def succeeded(self):
        return self.exit_code == 0

    @property
    def failed(self):
        return self.exit_code != 0

    def match(self, expected, stdout=None, stderr=None, log=None, regex=None):
        """
        :param str|re.Pattern expected: Message to find in self.logged
        :param bool|None stdout: Look at stdout (default: yes)
        :param bool|None stderr: Look at stderr (default: yes)
        :param bool|None log: Look at what was logged (default: no)
        :param int|bool|None regex: Specify whether 'expected' should be a regex
        :return Match|None: Match found, if any
        """
        if stdout is None and stderr is None and log is None:
            # By default, look at stdout/stderr only
            stdout = stderr = True
        assert expected, "No 'expected' provided"
        assert self.exit_code is not None, "run() was not called yet"
        captures = [stdout and self.logged.stdout, stderr and self.logged.stderr, log and self.logged.log]
        captures = [c for c in captures if c is not None and c is not False]

        assert captures, "No captures specified"
        if not any(c for c in captures):
            # There was no output at all
            return None

        if not isinstance(regex, bool) and isinstance(regex, int):
            flags = regex
            regex = True
        else:
            flags = 0

        if not isinstance(expected, string_type):
            # Assume regex, no easy way to verify isinstance(expected, re.Pattern) for python < 3.7
            regex = expected

        elif "..." in expected and regex is not False and regex is not True:
            regex = True
            expected = expected.replace("...", ".+")

        if regex is True:
            regex = re.compile("(.{0,32})(%s)(.{0,32})" % expected, flags=flags)

        for c in captures:
            contents = c.contents()
            if regex:
                m = regex.search(contents)
                if m:
                    if m.groups():
                        return Match(c, m.group(2), pre=m.group(1), post=m.group(3))
                    return Match(c, m.group(0))
            elif expected in contents:
                i = contents.index(expected)
                pre = runez.shortened(contents[:i], 32)
                post = runez.shortened(contents[i+len(expected):], 32)
                return Match(c, expected, pre=pre, post=post)

    def expect_messages(self, *expected, **kwargs):
        for message in expected:
            if message[0] == "!":
                assert not self.match(message[1:], **kwargs)
            else:
                assert self.match(message, **kwargs)

    def expect_success(self, args, *expected, **kwargs):
        self.run(args)
        assert self.succeeded
        self.expect_messages(*expected, **kwargs)

    def expect_failure(self, args, *expected, **kwargs):
        self.run(args)
        assert self.failed
        self.expect_messages(*expected, **kwargs)


class Match:
    def __init__(self, capture, match, pre=None, post=None):
        self.capture = capture
        self.match = match
        self.pre = pre
        self.post = post

    def __repr__(self):
        return self.match
